/*
 Standalone Temperature Logger - ATtiny85 Side
 Version: 1.0
 Author: Alex from Inside Gadgets (http://www.insidegadgets.com)
 Created: 9/01/2011
 Last Modified: 5/03/2011
 
 Used to monitor temperature every x minutes/hours which is saved 
 to the built in EEPROM. 
 
 */

#include <avr/sleep.h>
#include <avr/wdt.h>
#include <math.h>
#include <EEPROM.h>

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

volatile boolean f_pin = 0; // Pin change variable

int ledPin = 0;
int thermistorPin = 1; // Analog input 1
int buttonPin = 4;
int dataPin = 1;
int clockPin = 3; // Analog input 3

int functionSelect = 0;
int dataCount = 1;
int delayTime = -1;

void setup() {
  pinMode(ledPin, OUTPUT);
  pinMode(buttonPin, INPUT);  
  pinMode(dataPin, OUTPUT);
  
  // Pin interrupt setup
  sbi(GIMSK,PCIE); // Enable pin change interrupt 
  sbi(PCMSK,PCINT4); // Apply to pin D4 (button)
}

void loop() {

  // Wait for input
  if (functionSelect == 0) {
    
    // Read delay time
    if (delayTime == -1) {
      delayTime = EEPROM.read(0);
      if (delayTime > 10) { // Not a normal value
        delayTime = 0;
      }
    }
    
    // Always reset variables just in case
    digitalWrite(ledPin, LOW);
    digitalWrite(dataPin, LOW);

    system_sleep();
    
    // First test we are connected to the Arduino
    int clockState = analogRead(clockPin);
    if (clockState >= 160) {
      int clockhighCount = 0;
      // Delay for 10 x 25ms = 250ms, check CLK is always high
      for (int x = 0; x < 10; x++) { 
        delay(25);
        clockState = analogRead(clockPin);
        if (clockState >= 160) {
          clockhighCount++;
        }
        else {
          clockhighCount = 0;
        }
      }
      // CLK is high, now give the 250ms HIGH as an ok
      if (clockhighCount >= 10) {
        functionSelect = 3;
        digitalWrite(dataPin, HIGH);
        delay(250);
        digitalWrite(dataPin, LOW);
      }
    }
    else {
      // Button was pressed
      int buttonState = 0; 
      int buttonCount = 0;
      int buttonPressed = 0;
      long resetTimer = millis();
      while (millis() < resetTimer + 1500) {
        buttonState = digitalRead(buttonPin);
        if (buttonState == HIGH) {
          buttonCount++;
          while (buttonState == HIGH && millis() < resetTimer + 1500) {
            buttonState = digitalRead(buttonPin);
            buttonPressed++;
            delay(50);
          }
        }
      }
      
      if (buttonCount == 3) { // Start logging
        functionSelect = 1;
        dataCount = 1;
        for (int x = 0; x < 3; x++) {
          digitalWrite(ledPin, HIGH);
          delay(500);
          digitalWrite(ledPin, LOW);
          delay(500);
        }
      }
      else if (buttonPressed >= 20) { // Change delay time
        functionSelect = 2;
        delayTime = -1;
        digitalWrite(ledPin, HIGH);
        delay(2000);
        digitalWrite(ledPin, LOW);
      }
    }
  }
  
  // Log Temperature
  if (functionSelect == 1) {
    digitalWrite(dataPin, HIGH);
    int tempValue = int(Thermistor(analogRead(thermistorPin)));
    digitalWrite(dataPin, LOW);
    
    // Calculations to store -40C to 125C int to a single byte
    byte tempStore = tempValue + 130;
    if (tempValue <= -41) { // Temp is under min temp accepted
      tempStore = 90;
    }
    if (tempValue >= 126) { // Temp is over max temp accepted
      tempStore = 255;
    }
    
    // Write to EEPROM
    if (dataCount < 512) {
      EEPROM.write(dataCount, tempStore);
      dataCount++;
      if (dataCount < 512) {
        EEPROM.write(dataCount, 0); // So we know where to stop when extracting data
      }
    }
    else {
      functionSelect = 0;
    }
    
    // Calculate the delay time
    int waitCounter = 0;
    int waitTime = 0;
    f_pin = 0;
    if (delayTime == 0) {
      waitTime = 15;
    }
    if (delayTime >= 1 && delayTime < 4) {
      waitTime = (delayTime * 5) * 15;
    }
    if (delayTime >= 4 && delayTime < 6) {
      waitTime = ((delayTime-3) * 30) * 15;
    }
    if (delayTime >= 6 && delayTime < 8) {
      waitTime = ((delayTime-5) * 120) * 15;
    }
    if (delayTime == 8) {
      waitTime = 8 * 60 * 15;
    }
    if (delayTime == 9) {
      waitTime = 12 * 60 * 15;
    }
    if (delayTime == 10) {
      waitTime = 24 * 60 * 15;
    }
    
    while (waitCounter != waitTime) {
      setup_watchdog(8);
      system_sleep();
      waitCounter++;
      if (f_pin == 1) { // Did a pin change wake us up?
        functionSelect = 0;
        // Disable the watchdog
        MCUSR &= ~(1<<WDRF);
        WDTCR |= (1<<WDCE) | (1<<WDE);
        WDTCR = 0;
        
        waitCounter = waitTime; // break out of the loop
        for (int x = 0; x < 2; x++) {
          digitalWrite(ledPin, HIGH);
          delay(500);
          digitalWrite(ledPin, LOW);
          delay(500);
        }
      }
    }
  }
  
  // Configure the delay time
  if (functionSelect == 2) {
    int buttonState = digitalRead(buttonPin);
    if (buttonState == HIGH) {
      delay(500);
      int buttonPressed = false;
      buttonState = digitalRead(buttonPin);
      if (buttonState == HIGH && delayTime >= 0) {
        delay(1500);
        buttonState = digitalRead(buttonPin);
        if (buttonState == HIGH) {
          functionSelect = 0;
          EEPROM.write(0, delayTime);
          for (int x = 0; x < 3; x++) {
            digitalWrite(ledPin, HIGH);
            delay(500);
            digitalWrite(ledPin, LOW);
            delay(500);
          }
        }
        else {
          buttonPressed = true;
        }
      }
      if (buttonState == LOW || buttonPressed == true) {
        delayTime++;
        digitalWrite(ledPin, HIGH);
        delay(500);
        digitalWrite(ledPin, LOW);
        
        if (delayTime > 10) { // Too many presses
          delayTime = -1;
          digitalWrite(ledPin, HIGH);
          delay(1500);
          digitalWrite(ledPin, LOW);
        }
      }
    }
  }
  
  // Send data from EEPROM to Arduino
  if (functionSelect == 3) {
    digitalWrite(ledPin, HIGH);
    boolean inTransfer = true;
    int readCount = 1;
    while (inTransfer == true) {
      byte dataByte = 0;
      if (readCount < 512) {
        dataByte = EEPROM.read(readCount);
      }
      for (int x = 7; x >= 0; x--) {
        digitalWrite(dataPin, LOW);
        
        // Wait until CLK is low
        long resetTimer = millis();
        int clockIn = analogRead(clockPin);
        while (clockIn >= 160 && inTransfer == true) {
          clockIn = analogRead(clockPin);
          if (millis() > resetTimer + 250) {  // Check if host is restarting the CLK
            inTransfer = false;
            functionSelect = 0;
          }          
        }
        
        // Output the bit
        boolean dataBit = bitRead(dataByte, x);
        digitalWrite(dataPin, dataBit);
        
        // Wait until CLK is high
        resetTimer = millis();
        clockIn = analogRead(clockPin);
        while (clockIn == 0 && inTransfer == true) {
          clockIn = analogRead(clockPin);
          if (millis() > resetTimer + 100) {
            digitalWrite(dataPin, LOW); // Don't want to wait 250ms for this otherwise Arduino will think it's beginning from the start again
          }
          if (millis() > resetTimer + 250) { // Check if host is restarting the CLK
            inTransfer = false;
            functionSelect = 0;
          }
        }
      }
      if (dataByte == 0) { // No more data to send
        inTransfer = false;
        functionSelect = 0;
      }
      readCount++;
    }
  }
}

// Taken from http://www.arduino.cc/playground/ComponentLib/Thermistor2
double Thermistor(int RawADC) {
  double Temp;
  Temp = log(((10240000/RawADC) - 10000));
  Temp = 1 / (0.001129148 + (0.000234125 * Temp) + (0.0000000876741 * Temp * Temp * Temp));
  Temp = Temp - 273.15; // Convert Kelvin to Celcius
  return Temp;
}

// From http://interface.khm.de/index.php/lab/experiments/sleep_watchdog_battery/
void system_sleep() {
  pinMode(ledPin, INPUT);
  pinMode(dataPin, INPUT);
  
  cbi(ADCSRA,ADEN); // Switch Analog to Digital converter OFF
  set_sleep_mode(SLEEP_MODE_PWR_DOWN); // Set sleep mode
  sleep_mode(); // System sleeps here
  sbi(ADCSRA,ADEN);  // Switch Analog to Digital converter ON
  
  pinMode(ledPin, OUTPUT);
  pinMode(dataPin, OUTPUT);
}

// Watchdog timeout values
// 0=16ms, 1=32ms, 2=64ms, 3=128ms, 4=250ms, 5=500ms
// 6=1sec, 7=2sec, 8=4sec, 9=8sec
// From http://interface.khm.de/index.php/lab/experiments/sleep_watchdog_battery/
void setup_watchdog(int ii) {
  byte bb;
  int ww;
  if (ii > 9 ) ii=9;
  bb=ii & 7;
  if (ii > 7) bb|= (1<<5);
  bb|= (1<<WDCE);
  ww=bb;

  MCUSR &= ~(1<<WDRF);
  // Start timed sequence
  WDTCR |= (1<<WDCE) | (1<<WDE);
  // Set new watchdog timeout value
  WDTCR = bb;
  WDTCR |= _BV(WDIE);

}

ISR(WDT_vect) {
}

ISR(PCINT0_vect) {
  f_pin = 1; // Set global flag
}
